"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var code_generation_diagnostic_1 = require("../code-generation-diagnostic");
var debug = require("debug");
var LOG = debug("transform/speedyjs-transform-visitor");
/**
 * SpeedyJS transformer that invokes the WebAssembly Code generator when ever a SpeedyJS Entry function is declared in the
 * source code (an entry function is marked async and with the "use speedyjs" directive.
 */
var SpeedyJSTransformVisitor = (function () {
    function SpeedyJSTransformVisitor(compilationContext, codeGenerator) {
        this.compilationContext = compilationContext;
        this.codeGenerator = codeGenerator;
        this.inSpeedyJSFunction = false;
    }
    /**
     * Rewrites the source file in case any speedy js functions have been declared in this source file
     */
    SpeedyJSTransformVisitor.prototype.visitSourceFile = function (sourceFile, context) {
        this.codeGenerator.beginSourceFile(sourceFile, this.compilationContext, (function (emitHelper) { return context.requestEmitHelper(emitHelper); }));
        sourceFile = ts.updateSourceFileNode(sourceFile, context.visitLexicalEnvironment(sourceFile.statements));
        return this.codeGenerator.completeSourceFile(sourceFile);
    };
    /**
     * Tests if a 'normal' JavaScript function refers to a non entry SpeedyJS function.
     * If this is the case, an error is emitted.
     */
    SpeedyJSTransformVisitor.prototype.visitIdentifier = function (identifier, context) {
        if (this.inSpeedyJSFunction) {
            return identifier;
        }
        var symbol = this.compilationContext.typeChecker.getSymbolAtLocation(identifier);
        if (typeof symbol !== "undefined" && symbol.flags & ts.SymbolFlags.Function) {
            for (var _i = 0, _a = symbol.getDeclarations(); _i < _a.length; _i++) {
                var declaration = _a[_i];
                if (isSpeedyJSFunction(declaration) && !isSpeedyJSEntryFunction(declaration)) {
                    throw code_generation_diagnostic_1.CodeGenerationDiagnostic.referenceToNonSpeedyJSEntryFunctionFromJS(identifier, symbol);
                }
            }
        }
        return identifier;
    };
    /**
     * Tests if the given function is a speedy js entry function and in this case, triggers the web assembly code
     * generation for the given declaration.
     */
    SpeedyJSTransformVisitor.prototype.visitFunctionDeclaration = function (functionDeclaration, context) {
        return this.visitFunctionLikeDeclaration(functionDeclaration, context);
    };
    SpeedyJSTransformVisitor.prototype.visitMethodDeclaration = function (methodDeclaration, context) {
        return this.visitFunctionLikeDeclaration(methodDeclaration, context);
    };
    SpeedyJSTransformVisitor.prototype.visitConstructorDeclaration = function (constructorDeclaration, context) {
        return this.visitFunctionLikeDeclaration(constructorDeclaration, context);
    };
    SpeedyJSTransformVisitor.prototype.visitFunctionLikeDeclaration = function (functionLikeDeclaration, context) {
        var speedyJSFunction = isSpeedyJSFunction(functionLikeDeclaration);
        if (!speedyJSFunction && functionLikeDeclaration.body) {
            return context.visitEachChild(functionLikeDeclaration);
        }
        this.inSpeedyJSFunction = true;
        try {
            if (isSpeedyJSEntryFunction(functionLikeDeclaration)) {
                var name = getName(functionLikeDeclaration, this.compilationContext.typeChecker);
                LOG("Found SpeedyJS entry function " + name);
                validateSpeedyJSFunction(functionLikeDeclaration, this.compilationContext.typeChecker);
                return this.codeGenerator.generateEntryFunction(functionLikeDeclaration);
            }
            else {
                context.visitEachChild(functionLikeDeclaration);
                return ts.createNotEmittedStatement(functionLikeDeclaration); // remove the function
            }
        }
        finally {
            this.inSpeedyJSFunction = false;
        }
    };
    SpeedyJSTransformVisitor.prototype.fallback = function (node, context) {
        return context.visitEachChild(node);
    };
    return SpeedyJSTransformVisitor;
}());
exports.SpeedyJSTransformVisitor = SpeedyJSTransformVisitor;
/**
 * tests if the passed function is a speedy js function
 * @param fun the function to test
 * @return {boolean} true if the function is a speedy js function
 */
function isSpeedyJSFunction(fun) {
    if (!fun.body || !isBlock(fun.body)) {
        return false;
    }
    for (var _i = 0, _a = fun.body.statements; _i < _a.length; _i++) {
        var statement = _a[_i];
        if (isPrologueDirective(statement)) {
            if (statement.expression.text === "use speedyjs") {
                return true;
            }
        }
        else {
            break;
        }
    }
    return false;
}
/**
 * A speedy js function has the async modifier and contains "use speedyjs" directive
 * @param fun the function to test
 * @return {boolean} true if it is a speedy js entry function
 */
function isSpeedyJSEntryFunction(fun) {
    return isSpeedyJSFunction(fun) &&
        fun.kind === ts.SyntaxKind.FunctionDeclaration &&
        !!fun.modifiers &&
        !!fun.modifiers.find(function (modifier) { return modifier.kind === ts.SyntaxKind.AsyncKeyword; });
}
/**
 * Tests if the given function is a valid speedy js function.
 * A function is a valid SpeedyJS function if
 * * it has a name
 * * it has no optional arguments (is not overloaded)
 * * is not overloaded
 * * is not generic
 * @param fun the function to test
 * @param typeChecker the type checker
 * @throws if the function is not valid
 */
function validateSpeedyJSFunction(fun, typeChecker) {
    if (!fun.name) {
        throw code_generation_diagnostic_1.CodeGenerationDiagnostic.anonymousEntryFunctionsNotSupported(fun);
    }
    var optional = fun.parameters.find(function (parameter) { return !!parameter.questionToken || !!parameter.dotDotDotToken; });
    if (optional) {
        throw code_generation_diagnostic_1.CodeGenerationDiagnostic.optionalParametersInEntryFunctionNotSupported(optional);
    }
    if (fun.typeParameters && fun.typeParameters.length > 0) {
        throw code_generation_diagnostic_1.CodeGenerationDiagnostic.genericEntryFunctionNotSupported(fun);
    }
    if (typeChecker.isImplementationOfOverload(fun)) {
        throw code_generation_diagnostic_1.CodeGenerationDiagnostic.overloadedEntryFunctionNotSupported(fun);
    }
}
function getName(functionDeclaration, typeChecker) {
    if (!functionDeclaration.name) {
        return "**anonymous**";
    }
    var symbol = typeChecker.getSymbolAtLocation(functionDeclaration.name);
    return typeChecker.getFullyQualifiedName(symbol);
}
function isPrologueDirective(node) {
    return node.kind === ts.SyntaxKind.ExpressionStatement && node.expression.kind === ts.SyntaxKind.StringLiteral;
}
function isBlock(node) {
    return node.kind === ts.SyntaxKind.Block;
}
//# sourceMappingURL=speedyjs-transform-visitor.js.map